/*
 * File:
 *   crcp.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Cooperative remote copy.
 *
 * Copyright (c) 2004-2005.
 */

#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include <event.h>

#include "config.h"

#include "common.h"
#include "log.h"
#include "message.h"
#include "network.h"
#include "peer.h"
#include "protocol.h"
#include "source.h"
#include "util.h"

/* ################################################################### *
 * VARIABLES
 * ################################################################### */

static config_t config_;

/* Options */
static struct option long_options_[] = {
  /* These options set a flag */
  { "silent",   no_argument,       &config_.verbosity, LOG_SILENT },
  { "fatal",    no_argument,       &config_.verbosity, LOG_FATAL },
  { "error",    no_argument,       &config_.verbosity, LOG_ERROR },
  { "warn",     no_argument,       &config_.verbosity, LOG_WARN },
  { "info",     no_argument,       &config_.verbosity, LOG_INFO },
  { "debug",    no_argument,       &config_.verbosity, LOG_DEBUG },
  /* These options don't set a flag */
  { "program",  required_argument, 0, 'C' },
  { "host",     required_argument, 0, 'H' },
  { "port",     required_argument, 0, 'P' },
  { "ssh",      required_argument, 0, 'S' },
  { "timeout",  required_argument, 0, 'T' },
  { "version",  no_argument,       0, 'V' },
  { "daemon",   no_argument,       0, 'd' },
  { "help",     no_argument,       0, 'h' },
  { "protocol", required_argument, 0, 'p' },
  { "size",     required_argument, 0, 's' },
  { "extra",    required_argument, 0, 'x' },
  { 0, 0, 0, 0 }
};

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * usage
 */
static void usage()
{
  int i;

  logm(LOG_SILENT, "crcp -- cooperative remote file copy\n");
  logm(LOG_SILENT, "\n");
  logm(LOG_SILENT, "Usage:\n"
       "  crcp [options...] local remote...\n"
       "  crcp [options...] local @file\n"
       "  crcp [options...]\n");
  logm(LOG_SILENT, "\n");
  logm(LOG_SILENT, "Arguments:\n");
  logm(LOG_SILENT, "  local\n"
       "        Local path name of the form `%s<path>'\n", FILE_PREFIX);
  logm(LOG_SILENT, "  remote\n"
       "        Remote path name of the form `%s<user>@<host>:<port>/<path>' or `%s<host>:<port>/<path>'\n", SSH_PREFIX, TCP_PREFIX);
  logm(LOG_SILENT, "  @file\n"
       "        File containing a list of remote path names\n");
  logm(LOG_SILENT, "  <no argument>\n"
       "        Wait for incoming file\n");
  logm(LOG_SILENT, "\n");
  logm(LOG_SILENT, "Options:\n"
       "  --silent\n"
       "  --fatal\n"
       "  --error\n"
       "  --warn\n"
       "  --info\n"
       "  --debug\n"
       "        Print no, little, or much information (default=warn)\n");
  logm(LOG_SILENT, "  -C, --program <string> (default=\"%s\")\n"
       "        Path to the crcp program to start remotely using ssh\n", DEFAULT_PROGRAM);
  logm(LOG_SILENT, "  -H, --host <string> (default=\"%s\")\n"
       "        Local host\n", DEFAULT_LOCAL_HOST);
  logm(LOG_SILENT, "  -P, --port <int>\n"
       "        Local port (default=%d)\n", DEFAULT_LOCAL_PORT);
  logm(LOG_SILENT, "  -S, --ssh \"option\"\n"
       "        Option passed to ssh, e.g., -S \"-2\" (one ssh option per -S)\n");
  logm(LOG_SILENT, "  -T, --timeout <int>\n"
       "        Timeout before shutting down when there is no activity (default=%d seconds)\n", DEFAULT_TIMEOUT);
  logm(LOG_SILENT, "  -V, --version\n"
       "        Print version number\n");
  logm(LOG_SILENT, "  -d, --daemon\n"
       "        Run program as deamon (peer only)\n");
  logm(LOG_SILENT, "  -h, --help\n"
       "        Print this message\n");
  logm(LOG_SILENT, "  -p, --protocol <string>\n"
       "        Protocol to use for content distribution (default=\"%s\")\n", DEFAULT_PROTOCOL);
  logm(LOG_SILENT, "  -s, --size <int>\n"
       "        Define chunk size (default=%d bytes)\n", DEFAULT_CHUNK_SIZE);
  logm(LOG_SILENT, "  -x, --extra \"args\"\n"
       "        Extra protocol-specific arguments");
  logm(LOG_SILENT, "\n");
  logm(LOG_SILENT, "Valid protocols:\n");

  for (i = 0; ; i++) {
    protocol_t *p = get_protocol_by_id(i);
    if (p == NULL)
      break;
    logm(LOG_SILENT, "  %s\n", p->name);
  }
}

/* ======================================================================
 * signal_cb
 */
void signal_cb(int fd, short event, void *arg)
{
  logm(LOG_INFO, "Exiting on signal %d\n", fd);
  event_loopexit(NULL);
}

/* ======================================================================
 * main
 */
int main(int argc, char **argv)
{
  int i, is_source = 0, is_daemon = 0, log_fd = -1;
  char log_name[128];
  arg_list_t ssh_args;
  struct event sigterm_ev, sigint_ev, sighup_ev;

  config_.verbosity = DEFAULT_VERBOSITY;
  config_.local_port = DEFAULT_LOCAL_PORT;
  config_.local_host = DEFAULT_LOCAL_HOST;
  config_.chunk_size = DEFAULT_CHUNK_SIZE;
  config_.timeout = DEFAULT_TIMEOUT;
  config_.program = DEFAULT_PROGRAM;
  config_.protocol = DEFAULT_PROTOCOL;
  config_.args = "";

  ssh_args.list = NULL;
  add_args(&ssh_args, "ssh");
  add_args(&ssh_args, "-x");
  add_args(&ssh_args, "-o");
  add_args(&ssh_args, "ForwardAgent=no");
  add_args(&ssh_args, "-o");
  add_args(&ssh_args, "ClearAllForwardings=no");

  while (1) {
    int option_index = 0;
    int c = getopt_long(argc, argv, "C:H:P:S:Vdhp:s:x:", long_options_, &option_index);

    if (c == -1)
      break;

    if (c == 0 && long_options_[option_index].flag == 0)
      c = long_options_[option_index].val;

    switch (c) {
     case 0:
       /* Flag is automatically set */
       break;
     case 'C':
       config_.program = optarg;
       break;
     case 'H':
       config_.local_host = optarg;
       break;
     case 'P':
       config_.local_port = atoi(optarg);
       break;
     case 'S':
       add_args(&ssh_args, optarg);
       break;
     case 'T':
       config_.timeout = atoi(optarg);
       break;
     case 'V':
       logm(LOG_SILENT, "crcp %s\n", VERSION);
       exit(0);
     case 'd':
       is_daemon = 1;
       break;
     case 'h':
       usage();
       exit(0);
     case 'p':
       config_.protocol = optarg;
       break;
     case 's':
       config_.chunk_size = atoi(optarg);
       break;
     case 'x':
       config_.args = optarg;
       break;
     case '?':
       /* getopt_long already printed an error message */
       fatal(0, "Use -h or --help for help\n");
    }
  }

  if (argc - optind == 0) {
    /* Destination: wait for data */
    is_source = 0;
  } else {
    /* Source: send data */
    is_source = 1;

    if (argc - optind < 2)
      fatal(1, "Too few arguments\nUse -h or --help for help\n");
  }

  if (get_protocol_by_name(config_.protocol) == NULL)
    fatal(1, "Invalid protocol\nUse -h or --help for help\n");

  log_set_verbosity(config_.verbosity);

  if(config_.verbosity >= LOG_DEBUG)
    add_args(&ssh_args, "-v");

  if (MUST_LOG(LOG_INFO)) {
    int i;
    logm(LOG_INFO, "Command line:");
    for (i = 0; i < argc; i++)
      logm(LOG_NOPREFIX, " %s", argv[i]);
    logm(LOG_NOPREFIX, "\n");
  }

  if (!is_source && is_daemon) {
    if (config_.verbosity > DEFAULT_VERBOSITY) {
      /* Redirect stderr to log file */
      snprintf(log_name, sizeof(log_name), "/tmp/crcp-%s-%d.log", config_.local_host, config_.local_port);
      log_fd = open(log_name, O_WRONLY | O_CREAT | O_TRUNC, 0);
      if (log_fd < 0)
        xfatal(LOG_ERROR, "%s: open %s\n", __func__, log_name);
      if (fchmod(log_fd, S_IRUSR | S_IWUSR) < 0)
        xlogm(LOG_ERROR, "%s: fchmod\n", __func__);
      dup2(log_fd, STDERR_FILENO);
      if (log_fd > 2)
        close(log_fd);
      /* Redirect stdin and stdout to /dev/null */
      log_fd = open("/dev/null", O_RDWR, 0);
      if (log_fd < 0)
        xfatal(LOG_ERROR, "%s: open /dev/null\n", __func__);
      dup2(log_fd, STDIN_FILENO);
      dup2(log_fd, STDOUT_FILENO);
      if (log_fd > 2)
        close(log_fd);
    }
    if (daemon(1, config_.verbosity > DEFAULT_VERBOSITY ? 1 : 0) < 0)
      xfatal(LOG_ERROR, "%s: daemon\n", __func__);
  }

  event_init();

  signal_set(&sigint_ev, SIGINT, signal_cb, NULL);
  signal_add(&sigint_ev, NULL);
  signal_set(&sigterm_ev, SIGTERM, signal_cb, NULL);
  signal_add(&sigterm_ev, NULL);
  signal_set(&sighup_ev, SIGHUP, signal_cb, NULL);
  signal_add(&sighup_ev, NULL);

  init_watchdog(&config_);

  if (!is_source) {
    init_peer(&config_);
  } else {
    init_source(argv[optind], &argv[optind + 1], argc - optind - 1, &ssh_args, &config_);
  }

  logm(LOG_DEBUG, "Entering main loop\n");
  event_dispatch();
  logm(LOG_DEBUG, "Exiting main loop\n");

  if (!is_source) {
    /* Destination: wait for data */
    terminate_peer();
  } else {
    /* Source: send data */
    terminate_source();
  }

  terminate_watchdog();

  for (i = 0; i < ssh_args.num; i++)
    free(ssh_args.list[i]);
  free(ssh_args.list);

  logm(LOG_INFO, "Program end\n");

  return 0;
}
