/*
 * File:
 *   file.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Represents a file.
 *
 * Copyright (c) 2004-2005.
 */

#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "file.h"
#include "log.h"
#include "util.h"

/* ################################################################### *
 * VARIABLES
 * ################################################################### */

static unsigned char bit_pos[] = {
  0x80,
  0x40,
  0x20,
  0x10,
  0x08,
  0x04,
  0x02,
  0x01
};

static unsigned char inv_bit_pos[] = {
  (unsigned char)~0x80,
  (unsigned char)~0x40,
  (unsigned char)~0x20,
  (unsigned char)~0x10,
  (unsigned char)~0x08,
  (unsigned char)~0x04,
  (unsigned char)~0x02,
  (unsigned char)~0x01
};

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * seek_file
 */
static int seek_file(file_t *f, unsigned int chunk)
{
  off_t off;

  if (f->fd < 0) {
    logm(LOG_ERROR, "%s: Invalid file descriptor\n", __func__);
    return -1;
  }

  off = chunk * f->chunk_size;
  if (off >= f->file_size) {
    logm(LOG_ERROR, "%s: Invalid chunk number\n", __func__);
    return -1;
  }

  if (lseek(f->fd, off, SEEK_SET) != off) {
    xlogm(LOG_ERROR, "%s: lseek\n", __func__);
    return -1;
  }

  return off;
}

/* ======================================================================
 * open_file
 */
int open_file(file_t *f, int mode)
{
  int flags;
  struct stat stats;

  logm(LOG_DEBUG, "%s: Opening file: %s\n", __func__, f->file_name);

  if (mode == OPEN_READ)
    flags = O_RDONLY;
  else
    flags = O_RDWR | O_CREAT;

  f->fd = open(f->file_name, flags);
  if (f->fd < 0) {
    xlogm(LOG_ERROR, "%s: open\n", __func__);
    return -1;
  }

  if (mode == OPEN_READ) {
    if (fstat(f->fd, &stats) < 0) {
      xlogm(LOG_ERROR, "%s: fstat\n", __func__);
      close_file(f);
      return -1;
    }
    f->file_size = stats.st_size;
  } else {
    if (fchmod(f->fd, S_IRUSR | S_IWUSR) < 0) {
      xlogm(LOG_ERROR, "%s: fchmod\n", __func__);
      close_file(f);
      return -1;
    }
  }

  return f->fd;
}

/* ======================================================================
 * close_file
 */
void close_file(file_t *f)
{
  if (f->fd >= 0) {
    logm(LOG_DEBUG, "%s: Closing file: %s\n", __func__, f->file_name);
    close(f->fd);
    f->fd = 0;
  }
}

/* ======================================================================
 * read_chunk
 */
int read_chunk(file_t *f, unsigned int chunk, unsigned char *buf, int buflen)
{
  int len;

  if (seek_file(f, chunk) < 0)
    return -1;

  len = f->chunk_size;
  if (len > f->file_size - chunk * f->chunk_size)
    len = f->file_size - chunk * f->chunk_size;

  if (len > buflen) {
    logm(LOG_ERROR, "%s: Buffer too small (%d instead of %d)\n", __func__, buflen, len);
    return -1;
  }

  if (xread(f->fd, buf, len) != len) {
    logm(LOG_ERROR, "%s: xread\n", __func__);
    return -1;
  }

  return len;
}

/* ======================================================================
 * write_chunk
 */
int write_chunk(file_t *f, unsigned int chunk, unsigned char *buf, int buflen)
{
  int len;

  if (seek_file(f, chunk) < 0)
    return -1;

  len = f->file_size - chunk * f->chunk_size;
  if (len > f->chunk_size)
    len = f->chunk_size;

  if (len > buflen) {
    logm(LOG_ERROR, "%s: Buffer too small (%d instead of %d)\n", __func__, buflen, len);
    return -1;
  }

  if (xwrite(f->fd, buf, len) != len) {
    logm(LOG_ERROR, "%s: xwrite\n", __func__);
    return -1;
  }

  return len;
}

/* ======================================================================
 * create_stats
 */
int create_stats(file_stats_t *fs, file_t *f)
{
  fs->total_chunks = TOTAL_CHUNKS(f);
  fs->nb_chunks = 0;
  fs->chunk_mask = xcalloc(1, (fs->total_chunks + 7) / 8);
  return 0;
}

/* ======================================================================
 * free_stats
 */
void free_stats(file_stats_t *fs)
{
  free(fs->chunk_mask);
  fs->chunk_mask = NULL;
}

/* ======================================================================
 * set_chunk
 */
void set_chunk(file_stats_t *fs, unsigned int chunk)
{
  fs->chunk_mask[chunk / 8] |= bit_pos[chunk % 8];
}

/* ======================================================================
 * clear_chunk
 */
void clear_chunk(file_stats_t *fs, unsigned int chunk)
{
  fs->chunk_mask[chunk / 8] &= inv_bit_pos[chunk % 8];
}

/* ======================================================================
 * set_all
 */
void set_all(file_stats_t *fs)
{
  memset(fs->chunk_mask, 0xFF, (fs->total_chunks + 7) / 8);
}

/* ======================================================================
 * clear_all
 */
void clear_all(file_stats_t *fs)
{
  memset(fs->chunk_mask, 0, (fs->total_chunks + 7) / 8);
}

/* ======================================================================
 * has_chunk
 */
int has_chunk(file_stats_t *fs, unsigned int chunk)
{
  return (fs->chunk_mask[chunk / 8] & bit_pos[chunk % 8]) != 0;
}
