/*
 * File:
 *   message.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Message management.
 *
 * Copyright (c) 2004-2005.
 */

#include <rpc/rpc.h>

#include "common.h"
#include "message.h"

/* ################################################################### *
 * DECLARATIONS
 * ################################################################### */

static int xdr_msg_init(XDR *xdrs, msg_init_t *init);
static int xdr_msg_complete(XDR *xdrs, msg_complete_t *complete);
static int xdr_msg_send_chunk(XDR *xdrs, msg_send_chunk_t *send_chunk);
static int xdr_msg_request_chunk(XDR *xdrs, msg_request_chunk_t *request_chunk);
static int xdr_msg_connect_peers(XDR *xdrs, msg_connect_peers_t *connect_peers);
static int xdr_msg_response(XDR *xdrs, msg_response_t *response);
static int xdr_msg_shutdown(XDR *xdrs, msg_shutdown_t *shutdown);

/* ################################################################### *
 * VARIABLES
 * ################################################################### */

static struct xdr_discrim xdr_msg_discrim[] = {
  { MSG_TYPE_INIT, (xdrproc_t)xdr_msg_init },
  { MSG_TYPE_COMPLETE, (xdrproc_t)xdr_msg_complete },
  { MSG_TYPE_SEND_CHUNK, (xdrproc_t)xdr_msg_send_chunk },
  { MSG_TYPE_REQUEST_CHUNK, (xdrproc_t)xdr_msg_request_chunk },
  { MSG_TYPE_CONNECT_PEERS, (xdrproc_t)xdr_msg_connect_peers },
  { MSG_TYPE_RESPONSE, (xdrproc_t)xdr_msg_response },
  { MSG_TYPE_SHUTDOWN, (xdrproc_t)xdr_msg_shutdown },
  { __dontcare__, NULL }
};

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * xdr_msg_init
 */
static int xdr_msg_init(XDR *xdrs, msg_init_t *init)
{
  if (xdrs->x_op == XDR_DECODE) {
    init->file_name = NULL;
    init->args = NULL;
  }

  if (!xdr_string(xdrs, &init->file_name, MAX_PATH))
    return FALSE;
  if (!xdr_u_int(xdrs, &init->file_size))
    return FALSE;
  if (!xdr_u_int(xdrs, &init->chunk_size))
    return FALSE;
  if (!xdr_u_int(xdrs, &init->protocol))
    return FALSE;
  if (!xdr_string(xdrs, &init->args, MAX_ARGS))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg_complete
 */
static int xdr_msg_complete(XDR *xdrs, msg_complete_t *complete)
{
  if (!xdr_double(xdrs, &complete->total_time))
    return FALSE;
  if (!xdr_double(xdrs, &complete->chunk_time_average))
    return FALSE;
  if (!xdr_double(xdrs, &complete->chunk_time_variance))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg_send_chunk
 */
static int xdr_msg_send_chunk(XDR *xdrs, msg_send_chunk_t *send_chunk)
{
  if (!xdr_u_int(xdrs, &send_chunk->chunk_id))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg_request_chunk
 */
static int xdr_msg_request_chunk(XDR *xdrs, msg_request_chunk_t *request_chunk)
{
  if (!xdr_u_int(xdrs, &request_chunk->chunk_id))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg_connect_peers
 */
static int xdr_msg_peer_address(XDR *xdrs, msg_peer_address_t *peer_address)
{
  if (xdrs->x_op == XDR_DECODE)
    peer_address->host = NULL;

  if (!xdr_string(xdrs, &peer_address->host, MAX_HOST))
    return FALSE;
  if (!xdr_int(xdrs, &peer_address->port))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg_connect_peers
 */
static int xdr_msg_connect_peers(XDR *xdrs, msg_connect_peers_t *connect_peers)
{
  if (xdrs->x_op == XDR_DECODE)
    connect_peers->peers = NULL;

  if (!xdr_array(xdrs,
                 (char**)(&(connect_peers->peers)),
                 &(connect_peers->nb),
                 MAX_PEERS,
                 sizeof(msg_peer_address_t),
                 (xdrproc_t)xdr_msg_peer_address))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg_response
 */
static int xdr_msg_response(XDR *xdrs, msg_response_t *response)
{
  if (!xdr_int(xdrs, &response->status))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg_shutdown
 */
static int xdr_msg_shutdown(XDR *xdrs, msg_shutdown_t *shutdown)
{
  if (!xdr_int(xdrs, &shutdown->status))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * xdr_msg
 */
static int xdr_msg(XDR *xdrs, msg_t *msg)
{
  if (!xdr_u_long(xdrs, &msg->flux_id))
    return FALSE;
  if (!xdr_union(xdrs, (int *)&msg->type, (char *)&msg->h, xdr_msg_discrim, (xdrproc_t)xdr_void))
    return FALSE;
  if (!xdr_u_int(xdrs, &msg->l))
    return FALSE;

  return TRUE;
}

/* ======================================================================
 * msg2xdr
 */
int msg2xdr(msg_t *msg, u_char *buf, uint32_t *len)
{
  XDR xdrs;

  xdrmem_create(&xdrs, (char *)buf, *len, XDR_ENCODE);

  if (!xdr_msg(&xdrs, msg)) {
    xdr_destroy(&xdrs);
    return -1;
  }

  *len = xdr_getpos(&xdrs);

  xdr_destroy(&xdrs);

  return 0;
}

/* ======================================================================
 * xdr2msg
 */
int xdr2msg(msg_t *msg, u_char *buf, uint32_t len)
{
  XDR xdrs;
  int ret = 0;

  xdrmem_create(&xdrs, (char *)buf, len, XDR_DECODE);

  if (!xdr_msg(&xdrs, msg))
    ret = -1;

  xdr_destroy(&xdrs);

  return ret;
}
