/*
 * File:
 *   network.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Network functions.
 *
 * Copyright (c) 2004-2005.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "log.h"
#include "message.h"
#include "network.h"
#include "util.h"

/* ################################################################### *
 * DEFINES
 * ################################################################### */

#define IO_SOCKET_CLOSED        -2

#define IO_STATUS_INIT          0
#define IO_STATUS_SIZE          1
#define IO_STATUS_HEADER        2
#define IO_STATUS_DATA          3

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * io_read
 */
static void next_message(io_out_t *out)
{
  msg_t *m;

  if (out->msg != NULL) {
    m = out->msg;
    out->msg = out->msg->next;
    free(m->d);
    free(m);
    out->io.status = IO_STATUS_INIT;
  }
}

/* ======================================================================
 * io_read
 */
static ssize_t io_read(int fd, buffer_t *buf, unsigned char *data)
{
  ssize_t res;

  if (data == NULL)
    data = buf->data;
  res = read(fd, data + buf->pos, buf->missing);
  if (res == -1) {
    if(errno != EINTR && errno != EAGAIN) {
      xlogm(LOG_ERROR, "%s: read\n", __func__);
      return -1;
    }
  } else if (res == 0) {
    /* Socket closed */
    return IO_SOCKET_CLOSED;
  } else {
    buf->pos += res;
    buf->missing -= res;
  }

  return buf->pos;
}

/* ======================================================================
 * io_write
 */
static ssize_t io_write(int fd, buffer_t *buf, unsigned char *data)
{
  ssize_t res;

  if (data == NULL)
    data = buf->data;
  res = write(fd, data + buf->pos, buf->missing);
  if (res == -1) {
    if(errno != EINTR && errno != EAGAIN) {
      xlogm(LOG_ERROR, "%s: write\n", __func__);
      return -1;
    }
  } else if (res == 0) {
    /* Socket closed */
    return -2;
  } else {
    buf->pos += res;
    buf->missing -= res;
  }

  return buf->pos;
}

/* ======================================================================
 * receive_cb
 */
static void receive_cb(int fd, short event, void *arg)
{
  io_in_t *in = arg;
  int error = NET_ERROR;

  logm(LOG_DEBUG, "%s: Receiving data\n", __func__);

  if (in->io.status == IO_STATUS_INIT) {
    /* Initial state: beginning of message */
    in->io.status = IO_STATUS_SIZE;
    in->io.buf.pos = 0;
    in->io.buf.missing = sizeof(uint32_t);
  }

  if (in->io.buf.missing > 0) {
    int i;

    logm(LOG_DEBUG, "%s: Reading %d bytes\n", __func__, in->io.buf.missing);
    i = io_read(fd, &in->io.buf, (in->io.status == IO_STATUS_DATA ? in->msg->d : NULL));
    if (i < 0) {
      if (i == IO_SOCKET_CLOSED) {
        logm(LOG_DEBUG, "%s: Socket closed\n", __func__);
        error = NET_DISCONNECT;
      } else
        logm(LOG_ERROR, "%s: io_read\n", __func__);
      goto error;
    }
    logm(LOG_DEBUG, "%s: Read %d bytes\n", __func__, i);
  }

  if (in->io.buf.missing == 0) {
    if (in->io.status == IO_STATUS_SIZE) {
      uint32_t buflen;

      buflen = ntohl(*(uint32_t *)in->io.buf.data);
      if (buflen > in->io.buf.size) {
        logm(LOG_ERROR, "%s: Message too long: %u\n", __func__, buflen);
        goto error;
      }
      in->io.status = IO_STATUS_HEADER;
      in->io.buf.pos = 0;
      in->io.buf.missing = buflen;
    } else if (in->io.status == IO_STATUS_HEADER) {
      if (xdr2msg(in->msg, in->io.buf.data, in->io.buf.pos) < 0) {
        logm(LOG_ERROR, "%s: Invalid message\n", __func__);
        goto error;
      }
      in->io.status = IO_STATUS_DATA;
      in->io.buf.pos = 0;
      in->io.buf.missing = in->msg->l;
      if (in->msg->l > 0)
        in->msg->d = xrealloc(in->msg->d, in->msg->l);
    }
  }

  if (in->io.buf.missing == 0 && in->io.status == IO_STATUS_DATA) {
    if (in->fn(in->msg, in->arg, NET_OK) != 0)
      goto out; /* I/O structures have been deleted => exit directly */
    in->io.status = IO_STATUS_INIT;
  }

  /* Register event again */
  if (event_add(&in->io.event, NULL) < 0) {
    logm(LOG_ERROR, "%s: event_add\n", __func__);
    goto error;
  }
 out:
  return;

 error:
  in->fn(in->msg, in->arg, error);
}

/* ======================================================================
 * send_cb
 */
static void send_cb(int fd, short event, void *arg)
{
  io_out_t *out = arg;

  logm(LOG_DEBUG, "%s: Sending data\n", __func__);

  if (out->io.status == IO_STATUS_INIT) {
    uint32_t buflen, xbuflen;

    /* Initial state: beginning of message */
    out->io.status = IO_STATUS_HEADER;
    out->io.buf.pos = 0;

    buflen = out->io.buf.size - sizeof(xbuflen);

    if (msg2xdr(out->msg, out->io.buf.data + sizeof(xbuflen), &buflen) < 0) {
      logm(LOG_ERROR, "%s: Cannot encode message\n", __func__);
      goto error;
    }
    xbuflen = htonl(buflen);
    memcpy(out->io.buf.data, &xbuflen, sizeof(xbuflen));

    out->io.buf.missing = buflen + sizeof(xbuflen);
  }

  if (out->io.buf.missing > 0) {
    int i;

    logm(LOG_DEBUG, "%s: Sending %d bytes\n", __func__, out->io.buf.missing);
    i = io_write(fd, &out->io.buf, (out->io.status == IO_STATUS_DATA ? out->msg->d : NULL));
    if (i < 0) {
      logm(LOG_ERROR, "%s: io_write\n", __func__);
      goto error;
    }
    logm(LOG_DEBUG, "%s: Sent %d bytes\n", __func__, i);
  }

  if (out->io.buf.missing == 0) {
    if (out->io.status == IO_STATUS_HEADER) {
      out->io.status = IO_STATUS_DATA;
      out->io.buf.pos = 0;
      out->io.buf.missing = out->msg->l;
    }
  }

  if (out->io.buf.missing == 0 && out->io.status == IO_STATUS_DATA) {
    if (out->fn(out->msg, out->arg, NET_OK) != 0)
      goto out; /* I/O structures have been deleted => exit directly */
    next_message(out); /* Prepare next message */
  }

  if (out->msg != NULL) {
    /* Register event again */
    if (event_add(&out->io.event, NULL) < 0) {
      logm(LOG_ERROR, "%s: event_add\n", __func__);
      goto error;
    }
  }
 out:
  return;

 error:
  out->fn(out->msg, out->arg, NET_ERROR);
  next_message(out); /* Try to restore consistent state */
}

/* ======================================================================
 * connect_cb
 */
static void connect_cb(int fd, short event, void *arg)
{
  int sock;
  struct sockaddr sa;
  socklen_t salen = sizeof(sa);
  io_listener_t *l = arg;

  logm(LOG_DEBUG, "%s: Connection request\n", __func__);

  sock = accept(fd, &sa, &salen);
  if (sock < 0) {
    xlogm(LOG_ERROR, "%s: accept\n", __func__);
    goto out;
  }

  l->fn(sock, l->arg, NET_OK);

 out:
  /* Register event again */
  if (event_add(&l->event, NULL) < 0) {
    logm(LOG_ERROR, "%s: event_add\n", __func__);
  }
}

/* ======================================================================
 * timer_cb
 */
static void timer_cb(int fd, short event, void *arg)
{
  io_timer_t *t = arg;

  logm(LOG_DEBUG, "%s: Timer\n", __func__);

  t->fn(t->arg);
}

/* ======================================================================
 * open_socket
 */
int open_socket(const char *host, int port)
{
  int sock;
  struct hostent *he;
  struct sockaddr_in addr;

  if ((he = gethostbyname(host)) == NULL) {
    xlogm(LOG_ERROR, "%s: gethostbyname\n", __func__);
    return -1;
  }

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    xlogm(LOG_ERROR, "%s: socket\n", __func__);
    return -1;
  }

  if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
    xlogm(LOG_ERROR, "%s: fcntl\n", __func__);
    close_socket(sock);
    return -1;
  }

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(addr.sin_zero), 0, 8);
  if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0 && errno != EINPROGRESS) {
    xlogm(LOG_ERROR, "%s: connect\n", __func__);
    close_socket(sock);
    return -1;
  }

  return sock;
}

/* ======================================================================
 * bind_socket
 */
int bind_socket(const char *host, int port)
{
  int sock;
  int val;
  struct hostent *he;
  struct sockaddr_in addr;

  if ((he = gethostbyname(host)) == NULL) {
    xlogm(LOG_ERROR, "%s: gethostbyname\n", __func__);
    return -1;
  }

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    xlogm(LOG_ERROR, "%s: socket\n", __func__);
    return -1;
  }

  val = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) < 0) {
    xlogm(LOG_ERROR, "%s: setsockopt\n", __func__);
    close_socket(sock);
    return -1;
  }

  if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
    xlogm(LOG_ERROR, "%s: fcntl\n", __func__);
    close_socket(sock);
    return -1;
  }

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(addr.sin_zero), 0, 8);
  if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
    xlogm(LOG_ERROR, "%s: bind\n", __func__);
    close_socket(sock);
    return -1;
  }

  return sock;
}

/* ======================================================================
 * close_socket
 */
void close_socket(int sock)
{
  if (sock >= 0)
    close(sock);
}

/* ======================================================================
 * open_io_in
 */
int open_io_in(io_in_t *in, int fd, int (*fn)(msg_t *, void *, int), void *arg)
{
  assert(fd >= 0);

  in->msg = (msg_t *)xmalloc(sizeof(msg_t));
  in->msg->l = 0;
  in->msg->d = NULL;
  in->msg->next = NULL;

  in->fn = fn;
  in->arg = arg;
  in->io.fd = fd;
  in->io.status = IO_STATUS_INIT;
  in->io.buf.size = BUF_SIZE;
  in->io.buf.pos = 0;
  in->io.buf.missing = 0;
  in->io.buf.data = xmalloc(in->io.buf.size);

  event_set(&in->io.event, in->io.fd, EV_READ, receive_cb, in);
  if (event_add(&in->io.event, NULL) < 0) {
    logm(LOG_ERROR, "%s: event_add\n", __func__);
    free(in->io.buf.data);
    in->io.buf.data = NULL;
    return -1;
  }
  return 0;
}

/* ======================================================================
 * close_io_in
 */
void close_io_in(io_in_t* in)
{
  event_del(&in->io.event);
  close_socket(in->io.fd);
  in->io.fd = -1;
  free(in->io.buf.data);
  in->io.buf.data = NULL;
  if (in->msg != NULL) {
    free(in->msg->d);
    free(in->msg);
    in->msg = NULL;
  }
}

/* ======================================================================
 * open_io_out
 */
int open_io_out(io_out_t *out, int fd, int (*fn)(msg_t *, void *, int), void *arg)
{
  out->msg = NULL;
  out->fn = fn;
  out->arg = arg;
  out->io.fd = fd;
  out->io.status = IO_STATUS_INIT;
  out->io.buf.size = BUF_SIZE;
  out->io.buf.pos = 0;
  out->io.buf.missing = 0;
  out->io.buf.data = xmalloc(out->io.buf.size);

  return 0;
}

/* ======================================================================
 * close_io_out
 */
void close_io_out(io_out_t *out)
{
  msg_t *m;

  event_del(&out->io.event);
  close_socket(out->io.fd);
  out->io.fd = -1;
  free(out->io.buf.data);
  out->io.buf.data = NULL;

  while (out->msg != NULL) {
    m = out->msg;
    out->msg = out->msg->next;
    free(m->d);
    free(m);
  }
}

/* ======================================================================
 * open_io_listener
 */
int open_io_listener(io_listener_t *l, const char *host, int port, int (*fn)(int, void *, int), void *arg)
{
  logm(LOG_DEBUG, "%s: Open listener on %s:%d\n", __func__, host, port);

  l->fn = fn;
  l->arg = arg;
  l->fd = bind_socket(host, port);
  if (l->fd < 0) {
    logm(LOG_ERROR, "%s: bind_socket\n", __func__);
    return -1;
  }

  if (listen(l->fd, 10) < 0) {
    logm(LOG_ERROR, "%s: listen\n", __func__);
    close_socket(l->fd);
    return -1;
  }

  event_set(&l->event, l->fd, EV_READ, connect_cb, l);

  if (event_add(&l->event, NULL) < 0) {
    fatal(1, "%s: event_add\n", __func__);
    close_socket(l->fd);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * close_io_listener
 */
void close_io_listener(io_listener_t *l)
{
  close_socket(l->fd);
}

/* ======================================================================
 * init_io_timer
 */
void init_io_timer(io_timer_t *t)
{
  evtimer_set(&t->event, timer_cb, t);
}

/* ======================================================================
 * set_io_timer
 */
int set_io_timer(io_timer_t *t, struct timeval *tv)
{
  if (event_add(&t->event, tv) < 0) {
    fatal(1, "%s: event_add\n", __func__);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * clear_io_timer
 */
void clear_io_timer(io_timer_t *t)
{
  evtimer_del(&t->event);
}

/* ======================================================================
 * send_message
 */
int send_message(io_out_t *out, msg_t *msg)
{
  assert(out->io.fd >= 0);

  msg->next = NULL;

  if (out->msg == NULL) {
    /* No other message being sent */
    out->msg = msg;
    out->io.status = IO_STATUS_INIT;

    event_set(&out->io.event, out->io.fd, EV_WRITE, send_cb, out);
    if (event_pending(&out->io.event, EV_WRITE, NULL)) {
      logm(LOG_ERROR, "%s: Send already in progress\n", __func__);
      out->msg = NULL;
      return -1;
    }
    if (event_add(&out->io.event, NULL) < 0) {
      logm(LOG_ERROR, "%s: event_add\n", __func__);
      out->msg = NULL;
      return -1;
    }
  } else {
    /* Append message */
    msg_t *m = out->msg;
    while (m->next != NULL)
      m = m->next;
    m->next = msg;
  }

  return 0;
}

/* ======================================================================
 * add_args
 */
void add_args(arg_list_t *args, const char *fmt, ...)
{
  va_list ap;
  char buf[1024];
  int nalloc;

  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  va_end(ap);

  nalloc = args->nalloc;
  if (args->list == NULL) {
    nalloc = 32;
    args->num = 0;
  } else if (args->num + 2 >= nalloc)
    nalloc *= 2;

  args->list = (char **)xrealloc(args->list, nalloc * sizeof(char *));
  args->nalloc = nalloc;
  args->list[args->num++] = xstrdup(buf);
  args->list[args->num] = NULL;
}

/* ======================================================================
 * ssh_exec
 */
void ssh_exec(ssh_connection_t *ssh, arg_list_t *args, const char *cmd)
{
  int pin[2], pout[2], reserved[2];

  assert(ssh->host != NULL);

  logm(LOG_INFO, "Remote host: %s%s%s\n", ssh->user == NULL ? "" : ssh->user, ssh->user == NULL ? "" : "@", ssh->host);
  logm(LOG_INFO, "Executing: %s\n", cmd);

  /*
   * Reserve two descriptors so that the real pipes won't get
   * descriptors 0 and 1 because that will screw up dup2 below
   */
  pipe(reserved);

  /* Create a socket pair for communicating with ssh */
  if (pipe(pin) < 0)
    xfatal(1, "%s: pipe\n", __func__);
  if (pipe(pout) < 0)
    xfatal(1, "%s: pipe\n", __func__);

  /* Free the reserved descriptors */
  close(reserved[0]);
  close(reserved[1]);

  /* Fork a child to execute the command on the remote host using ssh */
  ssh->pid = fork();
  if (ssh->pid == 0) {
    /* Child */
    if (ssh->user != NULL) {
      add_args(args, "-l");
      add_args(args, "%s", ssh->user);
    }
    add_args(args, "%s", ssh->host);
    add_args(args, "%s", cmd);

    if (MUST_LOG(LOG_DEBUG)) {
      int i;
      logm(LOG_INFO, "Executing:");
      for (i = 0; i < args->num; i++)
        logm(LOG_NOPREFIX, " %s", args->list[i]);
      logm(LOG_NOPREFIX, "\n");
    }

    /* Redirect stdin/stdout to parent */
    close(pin[1]);
    close(pout[0]);
    dup2(pin[0], 0);
    dup2(pout[1], 1);
    close(pin[0]);
    close(pout[1]);
    execvp(args->list[0], args->list);
    fatal(1, "%s: execvp %s\n", __func__, args->list[0]);
  } else if (ssh->pid < 0)
    xfatal(1, "%s: fork\n", __func__);

  /* Parent: close the other side, and return the local side */
  close(pin[0]);
  ssh->fdout = pin[1];
  close(pout[1]);
  ssh->fdin = pout[0];
}
