/*
 * File:
 *   network.h
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Network functions.
 *
 * Copyright (c) 2004-2005.
 */

#ifndef NETWORK_H
#define NETWORK_H

#include <sys/types.h>

#include <event.h>

/* ################################################################### *
 * DEFINES
 * ################################################################### */

#define NET_OK                  0
#define NET_ERROR               1
#define NET_DISCONNECT          2

#define NET_CLOSED(conn)        ((conn)->io.fd < 0)

/* ################################################################### *
 * TYPEDEFS
 * ################################################################### */

struct msg;

typedef struct buffer {
  unsigned char *data;
  int size;
  int pos;
  int missing;
} buffer_t;

typedef struct io {
  int fd;
  int status;
  buffer_t buf;
  struct event event;
} io_t;

typedef struct io_in {
  io_t io;
  int (*fn)(struct msg *, void *, int);
  void *arg;
  struct msg *msg;
} io_in_t;

typedef struct io_out {
  io_t io;
  int (*fn)(struct msg *, void *, int);
  void *arg;
  struct msg *msg;
} io_out_t;

typedef struct io_listener {
  int fd;
  struct event event;
  int (*fn)(int, void *, int);
  void *arg;
} io_listener_t;

typedef struct io_timer {
  struct event event;
  int (*fn)(void *);
  void *arg;
} io_timer_t;

typedef struct arg_list {
  char **list;
  int num;
  int nalloc;
} arg_list_t;

typedef struct connection {
  io_in_t in;
  io_out_t out;
} connection_t;

typedef struct ssh_connection {
  char *user;
  char *host;
  char *cmd;
  int fdin;
  int fdout;
  pid_t pid;
} ssh_connection_t;

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

int open_socket(const char *host, int port);
int bind_socket(const char *host, int port);
void close_socket(int sock);

int open_io_in(io_in_t *in, int fd, int (*fn)(struct msg *, void *, int), void *arg);
void close_io_in(io_in_t *in);

int open_io_out(io_out_t *out, int fd, int (*fn)(struct msg *, void *, int), void *arg);
void close_io_out(io_out_t *out);

int open_io_listener(io_listener_t *l, const char *host, int port, int (*fn)(int, void *, int), void *arg);
void close_io_listener(io_listener_t *l);

void init_io_timer(io_timer_t *t);
int set_io_timer(io_timer_t *t, struct timeval *tv);
void clear_io_timer(io_timer_t *t);

int send_message(io_out_t *out, struct msg *msg);

void add_args(arg_list_t *args, const char *fmt, ...);

void ssh_exec(ssh_connection_t *ssh, arg_list_t *args, const char *cmd);

#endif /* NETWORK_H */
