/*
 * File:
 *   peer.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Represents a peer.
 *
 * Copyright (c) 2004-2005.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include <event.h>

#include "common.h"
#include "file.h"
#include "log.h"
#include "message.h"
#include "network.h"
#include "peer.h"
#include "protocol.h"
#include "util.h"

/* ################################################################### *
 * VARIABLES
 * ################################################################### */

static io_listener_t listener_;

struct timeval watchdog_timeout_;
static io_timer_t watchdog_;

static flux_t *flux_list_;
static int flux_list_size_;
static int nb_fluxes_;

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * add_flux
 */
static void add_flux(flux_t *flux)
{
  if (nb_fluxes_ == flux_list_size_) {
    flux_list_size_ *= 2;
    flux_list_ = (flux_t *)xrealloc(flux_list_, flux_list_size_ * sizeof(flux_t));
  }

  flux_list_[nb_fluxes_] = *flux;
  nb_fluxes_++;
}

/* ======================================================================
 * find_flux
 */
static flux_t *find_flux(unsigned long id)
{
  int i;

  for (i = 0; i < nb_fluxes_; i++) {
    if (flux_list_[i].file.flux_id == id) {
      return &flux_list_[i];
    }
  }
  return NULL;
}

/* ======================================================================
 * close_flux
 */
static void close_flux(flux_t *flux)
{
  flux->protocol->peer.terminate(flux);

  /* Cleanup */
  clear_io_timer(&flux->timer);
  close_file(&flux->file);
  free_stats(&flux->stats);
  free(flux->file.file_name);
  free(flux->args);
  /* Remove from array */
  assert (flux >= flux_list_ && flux < flux_list_ + nb_fluxes_);
  nb_fluxes_--;
  while (flux < flux_list_ + nb_fluxes_) {
    *flux = *(flux + 1);
    flux++;
  }
}

/* ======================================================================
 * close_peer
 */
static void close_peer(peer_t *peer)
{ 
  logm(LOG_DEBUG, "%s: Remove remote peer\n", __func__);

  if (peer->flux != NULL) {
    peer->flux->nb_peers--;
    /* Remove inactive fluxes */
    if (peer->flux->nb_peers == 0)
      close_flux(peer->flux);
  }

  /* Cleanup */
  close_io_in(&peer->conn.in);
  close_io_out(&peer->conn.out); /* Might try to close socket twice... */
  clear_io_timer(&peer->timer);
  free(peer);

  if (nb_fluxes_ == 0) {
    logm(LOG_INFO, "%s: No more fluxes\n", __func__);
    logm(LOG_INFO, "%s: Exiting process\n", __func__);
    event_loopexit(NULL);
  }
}

/* ======================================================================
 * timeout_flux_cb
 */
static int timeout_flux_cb(void *arg)
{
  flux_t *flux = (flux_t *)arg;

  logm(LOG_DEBUG, "%s: Timeout\n", __func__);

  reset_watchdog();

  return flux->protocol->peer.timeout(flux, NULL);
}

/* ======================================================================
 * timeout_peer_cb
 */
static int timeout_peer_cb(void *arg)
{
  peer_t *peer = (peer_t *)arg;
  flux_t *flux;

  logm(LOG_DEBUG, "%s: Timeout\n", __func__);

  reset_watchdog();

  flux = peer->flux;
  assert(flux != NULL);

  if (flux->protocol->peer.timeout(flux, peer) < 0) {
    if (flux->src == peer)
      flux->src = NULL;
    close_peer(peer);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * deliver_cb
 */
static int deliver_msg_cb(msg_t *msg, void *arg, int flags)
{
  flux_t *flux = NULL;
  peer_t *peer = (peer_t *)arg;

  reset_watchdog();

  if (flags == NET_OK && msg->type == MSG_TYPE_INIT) {
    /* Initial message sent from the source: flux does not exist yet */
    flux_t f;

    assert(find_flux(msg->flux_id) == NULL);

    /* Create flux */
    f.file.flux_id = msg->flux_id;
    f.file.file_name = msg->h.init.file_name;
    f.file.file_size = msg->h.init.file_size;
    f.file.chunk_size = msg->h.init.chunk_size;
    if (open_file(&f.file, OPEN_WRITE) < 0) {
      logm(LOG_ERROR, "%s: open_file\n", __func__);
      close_peer(peer);
      return -1;
    }
    if (create_stats(&f.stats, &f.file) < 0) {
      logm(LOG_ERROR, "%s: create_stats\n", __func__);
      close_file(&f.file);
      close_peer(peer);
      return -1;
    }
    f.protocol = get_protocol_by_id(msg->h.init.protocol);
    f.args = msg->h.init.args;
    add_flux(&f);
    flux = find_flux(msg->flux_id);
    assert(flux != NULL);

    assert(peer->flux == NULL);
    flux->src = peer;
    flux->nb_peers = 0;
    flux->private = NULL;

    flux->timer.fn = timeout_flux_cb;
    flux->timer.arg = flux;
    init_io_timer(&flux->timer);

    flux->protocol->peer.init(flux);
  }

  if (flux == NULL)
    flux = find_flux(msg->flux_id);

  if (flux == NULL) {
    /* Communication error before flux creation */
    assert(flags != NET_OK);
    logm(LOG_ERROR, "%s: Error while receiving initial message from remote peer\n", __func__);
    close_peer(peer);
    return -1;
  }

  if (peer->flux == NULL) {
    peer->flux = flux;
    flux->nb_peers++;
  } else
    assert(peer->flux == flux);

  if (flags != NET_OK) {
    if (flags == NET_DISCONNECT)
      logm(LOG_INFO, "%s: Connection closed\n", __func__);
    else
      logm(LOG_ERROR, "%s: Error while receiving message from remote peer\n", __func__);
    flux->protocol->peer.error(flux, peer, (flags == NET_DISCONNECT ? ERROR_DISCONNECT : ERROR_RECEIVE));
    if (flux->src == peer)
      flux->src = NULL;
    close_peer(peer);
    return -1;
  }

  logm(LOG_DEBUG, "%s: Received message from remote peer\n", __func__);

  if (msg->type != MSG_TYPE_INIT) { /* Discard init messages */
    if (flux->protocol->peer.deliver(flux, msg, peer) < 0) {
      if (flux->src == peer)
        flux->src = NULL;
      close_peer(peer);
      return -1;
    }
  }

  return 0;
}

/* ======================================================================
 * sent_msg_cb
 */
static int sent_msg_cb(msg_t *msg, void *arg, int flags)
{
  flux_t *flux;
  peer_t *peer = (peer_t *)arg;

  reset_watchdog();

  flux = find_flux(msg->flux_id);

  if (flux == NULL) {
    logm(LOG_ERROR, "%s: Invalid state\n", __func__);
    close_peer(peer);
    return -1;
  }

  if (flags == NET_OK) {
    logm(LOG_DEBUG, "%s: Sent message to remote peer\n", __func__);
    if (flux->protocol->peer.sent(flux, msg, peer) < 0) {
      if (flux->src == peer)
        flux->src = NULL;
      close_peer(peer);
      return -1;
    }
  } else {
    logm(LOG_ERROR, "%s: Error while sending message to remote peer\n", __func__);
    flux->protocol->peer.error(flux, peer, ERROR_SEND);
    if (flux->src == peer)
      flux->src = NULL;
    close_peer(peer);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * connect_peer_cb
 */
static int connect_peer_cb(int fd, void *arg, int flags)
{
  peer_t *peer;

  reset_watchdog();

  if (flags == NET_ERROR) {
    logm(LOG_ERROR, "%s: Error while listening to remote peers\n", __func__);
    return 0;
  }

  logm(LOG_DEBUG, "%s: Creating connections to remote peer\n", __func__);

  peer = xcalloc(1, sizeof(*peer));

  if (open_io_in(&peer->conn.in, fd, deliver_msg_cb, peer) < 0) {
    logm(LOG_ERROR, "%s: init_io_in\n", __func__);
    free(peer);
    return -1;
  }

  if (open_io_out(&peer->conn.out, fd, sent_msg_cb, peer) < 0) {
    logm(LOG_ERROR, "%s: init_io_out\n", __func__);
    close_io_in(&peer->conn.in);
    free(peer);
    return -1;
  }

  peer->flux = NULL;
  peer->private = NULL;

  peer->timer.fn = timeout_peer_cb;
  peer->timer.arg = peer;
  init_io_timer(&peer->timer);

  return 0;
}

/* ======================================================================
 * init_peer
 */
void init_peer(config_t *config)
{
  if (open_io_listener(&listener_, config->local_host, config->local_port, connect_peer_cb, NULL) < 0)
    fatal(1, "%s: Cannot open listener\n", __func__);

  flux_list_ = (flux_t *)xmalloc(FLUX_LIST_SIZE * sizeof(flux_t));
  flux_list_size_ = FLUX_LIST_SIZE;
  nb_fluxes_ = 0;
}

/* ======================================================================
 * terminate_peer
 */
void terminate_peer()
{
  int i;

  close_io_listener(&listener_);
  for (i = nb_fluxes_ - 1; i >= 0; i--)
    close_flux(&flux_list_[i]);
  free(flux_list_);
}

/* ======================================================================
 * create_peer
 */
peer_t *create_peer(const char *host, int port, flux_t *flux)
{
  int fd;
  peer_t *peer;

  fd = open_socket(host, port);
  if (fd < 0) {
    logm(LOG_ERROR, "%s: open_socket\n", __func__);
    return NULL;
  }

  peer = xcalloc(1, sizeof(*peer));

  if (open_io_in(&peer->conn.in, fd, deliver_msg_cb, peer) < 0) {
    logm(LOG_ERROR, "%s: init_io_in\n", __func__);
    free(peer);
    return NULL;
  }

  if (open_io_out(&peer->conn.out, fd, sent_msg_cb, peer) < 0) {
    logm(LOG_ERROR, "%s: init_io_out\n", __func__);
    close_io_in(&peer->conn.in);
    free(peer);
    return NULL;
  }

  flux->nb_peers++;
  peer->flux = flux;
  peer->private = NULL;

  peer->timer.fn = timeout_peer_cb;
  peer->timer.arg = peer;
  init_io_timer(&peer->timer);

  return peer;
}

/* ======================================================================
 * timeout_watchdog_cb
 */
static int timeout_watchdog_cb(void *arg)
{
  logm(LOG_INFO, "%s: Timeout: shutting down\n", __func__);
  event_loopexit(NULL);

  return 0;
}

/* ======================================================================
 * init_watchdog
 */
void init_watchdog(config_t *config)
{
  watchdog_.fn = timeout_watchdog_cb;
  watchdog_.arg = NULL;
  timerclear(&watchdog_timeout_);
  watchdog_timeout_.tv_sec = config->timeout;
  init_io_timer(&watchdog_);
  if (watchdog_timeout_.tv_sec > 0)
    set_io_timer(&watchdog_, &watchdog_timeout_);
}

/* ======================================================================
 * terminate_watchdog
 */
void terminate_watchdog()
{
  clear_io_timer(&watchdog_);
}

/* ======================================================================
 * reset_watchdog
 */
void reset_watchdog()
{
  if (watchdog_timeout_.tv_sec > 0)
    set_io_timer(&watchdog_, &watchdog_timeout_);
}
