/*
 * File:
 *   protocol.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Implements the distribution protocol.
 *
 * Copyright (c) 2004-2005.
 */

#include <assert.h>
#include <string.h>

#include "file.h"
#include "message.h"
#include "network.h"
#include "peer.h"
#include "protocol.h"
#include "protocol/parallel.h"
#include "protocol/linear.h"
#include "protocol/tree.h"
#include "protocol/ptree.h"
#include "log.h"
#include "stat.h"
#include "util.h"

/* ################################################################### *
 * VARIABLES
 * ################################################################### */

static protocol_t *protocols_[] = {
  &protocol_parallel,
  &protocol_linear,
  &protocol_tree,
  &protocol_ptree,
  NULL
};

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * get_protocol_id
 */
static int get_protocol_id(protocol_t *protocol)
{
  int i;

  for (i = 0; protocols_[i] != NULL; i++)
    if (protocol == protocols_[i])
      return i;

  return -1;
}

/* ======================================================================
 * get_protocol_by_name
 */
protocol_t *get_protocol_by_name(const char *name)
{
  int i;

  for (i = 0; protocols_[i] != NULL; i++)
    if (strcmp(name, protocols_[i]->name) == 0)
      return protocols_[i];

  return NULL;
}

/* ======================================================================
 * get_protocol_by_id
 */
protocol_t *get_protocol_by_id(int id)
{
  int i;

  /* Make sure index is valid */
  for (i = 0; protocols_[i] != NULL; i++)
    if (i == id)
      return protocols_[i];

  return NULL;
}

/* ======================================================================
 * msg_init
 */
int msg_init(peer_t *peer, file_t *f, const char *path, protocol_t *protocol, const char *args)
{
  msg_t *msg;

  if (peer == NULL || NET_CLOSED(&peer->conn.out))
    return -1;

  msg = (msg_t *)xmalloc(sizeof(msg_t));

  msg->type = MSG_TYPE_INIT;
  msg->flux_id = f->flux_id;
  msg->h.init.file_name = (char *)path;
  msg->h.init.file_size = f->file_size;
  msg->h.init.chunk_size = f->chunk_size;
  msg->h.init.protocol = get_protocol_id(protocol);
  msg->h.init.args = (char *)args;
  if (msg->h.init.protocol < 0) {
    logm(LOG_ERROR, "%s: Invalid protocol: %d\n", __func__, msg->h.init.protocol);
    free(msg);
    return -1;
  }
  msg->d = NULL;
  msg->l = 0;
  if (send_message(&peer->conn.out, msg) < 0 ) {
    logm(LOG_ERROR, "%s: send_message\n", __func__);
    free(msg);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * msg_complete
 */
int msg_complete(peer_t *peer, file_t *f, double total_time, struct samples *chunk_time)
{
  msg_t *msg;

  if (peer == NULL || NET_CLOSED(&peer->conn.out))
    return -1;

  msg = (msg_t *)xmalloc(sizeof(msg_t));

  msg->type = MSG_TYPE_COMPLETE;
  msg->flux_id = f->flux_id;
  msg->h.complete.total_time = total_time;
  msg->h.complete.chunk_time_average = samples_mean(chunk_time);
  msg->h.complete.chunk_time_variance = samples_variance(chunk_time);
  msg->d = NULL;
  msg->l = 0;
  if (send_message(&peer->conn.out, msg) < 0 ) {
    logm(LOG_ERROR, "%s: send_message\n", __func__);
    free(msg);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * msg_send_chunk
 */
int msg_send_chunk(peer_t *peer, file_t *f, unsigned int chunk)
{
  int len;
  msg_t *msg;

  if (peer == NULL || NET_CLOSED(&peer->conn.out))
    return -1;

  msg = (msg_t *)xmalloc(sizeof(msg_t));

  msg->d = xmalloc(f->chunk_size);
  len = read_chunk(f, chunk, msg->d, f->chunk_size);
  if (len < 0) {
    logm(LOG_ERROR, "%s: Cannot read chunk from file\n", __func__);
    free(msg->d);
    free(msg);
    return -1;
  }
  msg->l = len;

  msg->type = MSG_TYPE_SEND_CHUNK;
  msg->flux_id = f->flux_id;
  msg->h.send_chunk.chunk_id = chunk;
  if (send_message(&peer->conn.out, msg) < 0 ) {
    logm(LOG_ERROR, "%s: send_message\n", __func__);
    free(msg->d);
    free(msg);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * msg_request_chunk
 */
int msg_request_chunk(peer_t *peer, file_t *f, unsigned int chunk)
{
  msg_t *msg;

  if (peer == NULL || NET_CLOSED(&peer->conn.out))
    return -1;

  msg = (msg_t *)xmalloc(sizeof(msg_t));

  msg->type = MSG_TYPE_REQUEST_CHUNK;
  msg->flux_id = f->flux_id;
  msg->h.request_chunk.chunk_id = chunk;
  msg->d = NULL;
  msg->l = 0;
  if (send_message(&peer->conn.out, msg) < 0 ) {
    logm(LOG_ERROR, "%s: send_message\n", __func__);
    free(msg);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * msg_connect_peers
 */
int msg_connect_peers(peer_t *peer, file_t *f, msg_peer_address_t *peers, int nb)
{
  msg_t *msg;

  if (peer == NULL || NET_CLOSED(&peer->conn.out))
    return -1;

  msg = (msg_t *)xmalloc(sizeof(msg_t));

  msg->type = MSG_TYPE_CONNECT_PEERS;
  msg->flux_id = f->flux_id;
  msg->h.connect_peers.nb = nb;
  msg->h.connect_peers.peers = peers;
  msg->d = NULL;
  msg->l = 0;
  if (send_message(&peer->conn.out, msg) < 0 ) {
    logm(LOG_ERROR, "%s: send_message\n", __func__);
    free(msg);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * msg_response
 */
int msg_response(peer_t *peer, file_t *f, int status)
{
  msg_t *msg;

  if (peer == NULL || NET_CLOSED(&peer->conn.out))
    return -1;

  msg = (msg_t *)xmalloc(sizeof(msg_t));

  msg->type = MSG_TYPE_RESPONSE;
  msg->flux_id = f->flux_id;
  msg->h.response.status = status;
  msg->d = NULL;
  msg->l = 0;
  if (send_message(&peer->conn.out, msg) < 0 ) {
    logm(LOG_ERROR, "%s: send_message\n", __func__);
    free(msg);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * msg_shutdown
 */
int msg_shutdown(peer_t *peer, file_t *f, int status)
{
  msg_t *msg;

  if (peer == NULL || NET_CLOSED(&peer->conn.out))
    return -1;

  msg = (msg_t *)xmalloc(sizeof(msg_t));

  msg->type = MSG_TYPE_SHUTDOWN;
  msg->flux_id = f->flux_id;
  msg->h.shutdown.status = status;
  msg->d = NULL;
  msg->l = 0;
  if (send_message(&peer->conn.out, msg) < 0 ) {
    logm(LOG_ERROR, "%s: send_message\n", __func__);
    free(msg);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * store_chunk
 */
int store_chunk(msg_t *msg, file_t *f, unsigned int chunk)
{
  if (write_chunk(f, chunk, msg->d, msg->l) < 0) {
    logm(LOG_ERROR, "%s: Cannot wite chunk to file\n", __func__);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * copy_level
 */
static int tree_bf2df_level(int *src, int *dst, int nb, int degree, int level, int size, int *idx)
{
  int i, n, cnt;
  int *sizes;

  cnt = 0;
  if (nb == 0) {
    /* Do nothing */
  } else if (level == 0) {
    dst[0] = src[(*idx)++ % size];
    cnt++;
  } else {
    sizes = tree_partition(nb - 1, degree);
    n = 1;
    for (i = 0; i < degree; i++) {
      cnt += tree_bf2df_level(src, &dst[n], sizes[i], degree, level - 1, size, idx);
      n += sizes[i];
    }
    free(sizes);
  }
  return cnt;
}

/* ======================================================================
 * tree_bf2df
 */
int *tree_bf2df(int nb, int degree, int start)
{
  int i, cnt, idx;
  int *src, *dst;

  src = (int *)xmalloc(nb * sizeof(int));
  dst = (int *)xmalloc(nb * sizeof(int));

  for (i = 0; i < nb; i++)
    src[i] = i;

  cnt = 0;
  idx = start;
  /* Traverse the tree several times to copy nodes one level after the
   * other (could be improved with breadth-first traversal using a
   * queue!) */
  for (i = 0; cnt < nb; i++)
    cnt += tree_bf2df_level(src, dst, nb, degree, i, nb, &idx);

  free(src);
  return dst;
}

/* ======================================================================
 * tree_depth
 */
static int tree_depth(int nb, int degree)
{
  int depth, n, level;

  depth = 0;
  n = 0;
  level = 1;
  while (n < nb) {
    n += level;
    level *= degree;
    depth++;
  }
  return depth;
}

/* ======================================================================
 * tree_size
 */
static int tree_size(int depth, int degree)
{
  int i, nb, level;

  nb = 0;
  level = 1;
  for (i = 0; i < depth; i++) {
    nb += level;
    level *= degree;
  }
  return nb;
}

/* ======================================================================
 * tree_partition
 */
int *tree_partition(int nb, int degree)
{
  int i, n, d, min, max;
  int *sizes;

  /* Compute the number of nodes in each partition */
  assert(degree > 0);
  sizes = (int *)xmalloc(degree * sizeof(int));

  i = degree - 1;
  n = (nb + degree - 1) / degree; /* ceil(nb/degree) */
  d = tree_depth(n, degree);
  max = tree_size(d, degree);
  min = tree_size(d - 1, degree);
  assert(min * degree <= nb && nb <= max * degree);
  while (i >= 0 && nb - (i * max) <= min)
    nb -= (sizes[i--] = min);
  while (i >= 0)
    nb -= (sizes[i--] = nb - (i * max));
  assert (nb == 0);

  return sizes;
}
