/*
 * File:
 *   protocol/parallel.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Implements the parallel distribution protocol.
 *
 * Copyright (c) 2004-2005.
 */

#include <assert.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>

#include <event.h>

#include "common.h"
#include "file.h"
#include "log.h"
#include "message.h"
#include "network.h"
#include "peer.h"
#include "protocol.h"
#include "protocol/parallel.h"
#include "source.h"
#include "stat.h"
#include "util.h"

/* ################################################################### *
 * TYPEDEFS
 * ################################################################### */

typedef struct source_data {
  samples_t samples;
  destination_t **dst;
  int nb_dst;
} source_data_t;

typedef struct peer_data {
  struct timeval start;
  struct timeval previous;
  samples_t samples;
  double total;
} peer_data_t;

/* ################################################################### *
 * DECLARATIONS
 * ################################################################### */

static void source_init(flux_t *flux, destination_t **dst, int nb_dst);
static void source_start(flux_t *flux);
static void source_terminate(flux_t *flux);
static void source_error(flux_t *flux, destination_t *dst, int type);
static int source_deliver(flux_t *flux, msg_t *msg, destination_t *dst);
static int source_sent(flux_t *flux, msg_t *msg, destination_t *dst);
static int source_timeout(flux_t *flux, destination_t *dst);

static void peer_init(flux_t *flux);
static void peer_terminate(flux_t *flux);
static void peer_error(flux_t *flux, peer_t *peer, int type);
static int peer_deliver(flux_t *flux, msg_t *msg, peer_t *peer);
static int peer_sent(flux_t *flux, msg_t *msg, peer_t *peer);
static int peer_timeout(flux_t *flux, peer_t *peer);

/* ################################################################### *
 * VARIABLES
 * ################################################################### */

protocol_t protocol_parallel = {
  "parallel",
  {
    source_init,
    source_start,
    source_terminate,
    source_error,
    source_deliver,
    source_sent,
    source_timeout
  },
  {
    peer_init,
    peer_terminate,
    peer_error,
    peer_deliver,
    peer_sent,
    peer_timeout
  }
};

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * source_init
 */
static void source_init(flux_t *flux, destination_t **dst, int nb_dst)
{
  source_data_t *data;

  logm(LOG_INFO, "%s: Initialize source\n", __func__);

  flux->private = xmalloc(sizeof(peer_data_t));
  data = (source_data_t *)flux->private;
  init_samples(&data->samples);

  data->dst = dst;
  data->nb_dst = nb_dst;
}

/* ======================================================================
 * source_start
 */
static void source_start(flux_t *flux)
{
  int i;
  source_data_t *data = (source_data_t *)flux->private;

  logm(LOG_INFO, "%s: Start source\n", __func__);

  for (i = 0; i < data->nb_dst; i++)
    msg_send_chunk(&data->dst[i]->peer, &flux->file, 0);
}

/* ======================================================================
 * flux_terminate
 */
static void source_terminate(flux_t *flux)
{
  source_data_t *data = (source_data_t *)flux->private;

  logm(LOG_INFO, "%s: Terminate source\n", __func__);

  if (MUST_LOG(LOG_INFO)) {
    logm(LOG_INFO, "\t# peers complete : %d\n", samples_count(&data->samples));
    if (samples_count(&data->samples) > 0) {
      logm(LOG_INFO, "\tPeer average     : %f (%f kB/s)\n", samples_mean(&data->samples),
           (double)flux->file.file_size / samples_mean(&data->samples) / 1024);
      logm(LOG_INFO, "\tPeer variance    : %f\n", samples_variance(&data->samples));
    }
  }

  free(flux->private);
}

/* ======================================================================
 * source_error
 */
static void source_error(flux_t *flux, destination_t *dst, int type)
{
  source_data_t *data = (source_data_t *)flux->private;

  if (type != ERROR_DISCONNECT)
    logm(LOG_ERROR, "%s: Error on connection with %s:%d\n", __func__, dst->host, dst->port);

  data->nb_dst--;
}

/* ======================================================================
 * source_deliver
 */
static int source_deliver(flux_t *flux, msg_t *msg, destination_t *dst)
{
  source_data_t *data = (source_data_t *)flux->private;

  switch(msg->type) {

   case MSG_TYPE_COMPLETE:
     logm(LOG_INFO, "%s: %s:%d COMPLETE: %f (%f kB/s) - avg: %f - var: %f\n", __func__,
          dst->host, dst->port,
          msg->h.complete.total_time,
          (double)flux->file.file_size / msg->h.complete.total_time / 1024,
          msg->h.complete.chunk_time_average,
          msg->h.complete.chunk_time_variance);
     add_sample(&data->samples, msg->h.complete.total_time);

     msg_shutdown(&dst->peer, &flux->file, STATUS_OK);
     break;

   default:
    logm(LOG_ERROR, "%s: Invalid message type: %d\n", __func__, msg->type);
    break;
  }

  return 0;
}

/* ======================================================================
 * source_sent
 */
static int source_sent(flux_t *flux, msg_t *msg, destination_t *dst)
{
  switch(msg->type) {

   case MSG_TYPE_SEND_CHUNK:
     if (msg->h.send_chunk.chunk_id + 1 < TOTAL_CHUNKS(&flux->file))
       msg_send_chunk(&dst->peer, &flux->file, msg->h.send_chunk.chunk_id + 1);
     else
       logm(LOG_INFO, "%s: Finished sending file to: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);
     break;

   case MSG_TYPE_SHUTDOWN:
     break;

   default:
    logm(LOG_ERROR, "%s: Invalid message type: %d\n", __func__, msg->type);
    break;
  }

  return 0;
}

/* ======================================================================
 * flux_timeout
 */
static int source_timeout(flux_t *flux, destination_t *dst)
{
  return 0;
}

/* ======================================================================
 * peer_init
 */
static void peer_init(flux_t *flux)
{
  peer_data_t *data;

  logm(LOG_INFO, "%s: Initialize peer\n", __func__);

  flux->private = xmalloc(sizeof(peer_data_t));
  data = (peer_data_t *)flux->private;
  init_samples(&data->samples);

  gettimeofday(&data->start, NULL);
  data->previous = data->start;
}

/* ======================================================================
 * peer_terminate
 */
static void peer_terminate(flux_t *flux)
{
  logm(LOG_INFO, "%s: Terminate peer\n", __func__);

  free(flux->private);
}

/* ======================================================================
 * peer_error
 */
static void peer_error(flux_t *flux, peer_t *peer, int type)
{
  if (type != ERROR_DISCONNECT)
    logm(LOG_ERROR, "%s: Error on connection with remote peer\n", __func__);
}

/* ======================================================================
 * peer_deliver
 */
static int peer_deliver(flux_t *flux, msg_t *msg, peer_t *peer)
{
  struct timeval tv;
  double d;
  peer_data_t *data = (peer_data_t *)flux->private;

  switch(msg->type) {

   case MSG_TYPE_SEND_CHUNK:
     logm(LOG_INFO, "%s: SEND CHUNK: %u\n", __func__, msg->h.send_chunk.chunk_id);
     gettimeofday(&tv, NULL);

     d = (double)tv.tv_sec + (double)tv.tv_usec / 1000000;
     d -= (double)data->previous.tv_sec + (double)data->previous.tv_usec / 1000000;
     add_sample(&data->samples, d);
     data->previous = tv;

     if (msg->h.send_chunk.chunk_id > flux->stats.total_chunks) {
       logm(LOG_ERROR, "%s: Invalid chunk ID: %u\n", __func__, msg->h.send_chunk.chunk_id);
       return -1; /* Will close endpoint */
     }
     if (msg->h.send_chunk.chunk_id != flux->stats.nb_chunks) {
       logm(LOG_ERROR, "%s: Unexpected chunk ID: %u\n", __func__, msg->h.send_chunk.chunk_id);
       return -1; /* Will close endpoint */
     }
     if (has_chunk(&flux->stats, msg->h.send_chunk.chunk_id)) {
       logm(LOG_ERROR, "%s: Chunk already received: %u\n", __func__, msg->h.send_chunk.chunk_id);
       return -1; /* Will close endpoint */
     }

     if (store_chunk(msg, &flux->file, msg->h.send_chunk.chunk_id) < 0) {
       logm(LOG_ERROR, "%s: Cannot store chunk: %u\n", __func__, msg->h.send_chunk.chunk_id);
       return -1; /* Will close endpoint */
     }

     set_chunk(&flux->stats, msg->h.send_chunk.chunk_id);
     flux->stats.nb_chunks++;

     if (flux->stats.nb_chunks == flux->stats.total_chunks) {
       /* Done! */
       logm(LOG_INFO, "%s: File complete: %s\n", __func__, flux->file.file_name);

       d = (double)tv.tv_sec + (double)tv.tv_usec / 1000000;
       d -= (double)data->start.tv_sec + (double)data->start.tv_usec / 1000000;
       data->total = d;

       if (MUST_LOG(LOG_INFO)) {
         logm(LOG_INFO, "\tTotal time     : %f\n", data->total);
         logm(LOG_INFO, "\tChunk average  : %f\n", samples_mean(&data->samples));
         logm(LOG_INFO, "\tChunk variance : %f\n", samples_variance(&data->samples));
       }

       if (flux->src != NULL)
         msg_complete(flux->src, &flux->file, data->total, &data->samples);
     }
     break;

   case MSG_TYPE_SHUTDOWN:
     logm(LOG_INFO, "%s: SHUTDOWN: %d\n", __func__, msg->h.shutdown.status);
     return -1; /* Will close endpoint */

   default:
     logm(LOG_ERROR, "%s: Invalid message type: %d\n", __func__, msg->type);
     break;
  }

  return 0;
}

/* ======================================================================
 * peer_sent
 */
static int peer_sent(flux_t *flux, msg_t *msg, peer_t *peer)
{
  switch(msg->type) {

   case MSG_TYPE_COMPLETE:
    break;

   default:
     logm(LOG_ERROR, "%s: Invalid message type: %d\n", __func__, msg->type);
     break;
  }

  return 0;
}

/* ======================================================================
 * peer_timeout
 */
static int peer_timeout(flux_t *flux, peer_t *peer)
{
  return 0;
}
