/*
 * File:
 *   source.c
 * Author(s):
 *   Pascal Felber <Pascal.Felber@unine.ch>
 * Description:
 *   Represents a data source.
 *
 * Copyright (c) 2004-2005.
 */

#include <assert.h>
#include <crypt.h>
#include <ctype.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/wait.h>

#include <event.h>

#include "common.h"
#include "file.h"
#include "log.h"
#include "peer.h"
#include "message.h"
#include "network.h"
#include "protocol.h"
#include "source.h"
#include "util.h"

/* ################################################################### *
 * VARIABLES
 * ################################################################### */

static flux_t flux_;
static destination_t **dst_list_ = NULL;
static int nb_dst_ = 0;

/* ################################################################### *
 * FUNCTIONS
 * ################################################################### */

/* ======================================================================
 * parse_source
 */
static void parse_source(flux_t *flux, const char *s)
{
  const char *start;
  unsigned int len;

  if (strstr(s, FILE_PREFIX) != s)
    fatal(1, "Invalid local path: %s\n", s);

  start = s + strlen(FILE_PREFIX);
  len = strlen(start);
  if (len == 0 || start[len - 1] == '/')
    fatal(1, "Path must be fully specified (including file name): %s\n", s);
  else
    flux->file.file_name = xstrdup(start);
}

/* ======================================================================
 * parse_destination
 */
static void parse_destination(destination_t *dst, const char *s)
{
  const char *start = s, *end;
  unsigned int i, len;

  dst->ssh = NULL;
  dst->host = NULL;
  dst->port = 0;
  dst->path = NULL;

  if (strstr(s, SSH_PREFIX) == s) {
    /* Remote path, process started with ssh */
    dst->ssh = (ssh_connection_t *)xmalloc(sizeof(ssh_connection_t));
    start = s + strlen(SSH_PREFIX);
    end = strchr(start, '@');
    if (end != NULL) {
      /* ``user'' has been specified */
      len = end - start;
      /* Is username valid? */
      for(i = 0; i < len; i++) {
        if (!isalpha(start[i]) && !isdigit(start[i]) && start[i] != '_' && start[i] != '-')
          fatal(1, "Invalid user name in path: %s\n", s);
      };
      dst->ssh->user = (char *)xmalloc(len + 1);
      strncpy(dst->ssh->user, start, len);
      dst->ssh->user[len] = '\0';
      start = end + 1;
    } else
      dst->ssh->user = NULL;
    end = strchr(start, ':');
    if (end == NULL)
      fatal(1, "Invalid path: %s\n", s);
    len = end - start;
    dst->host = (char *)xmalloc(len + 1);
    strncpy(dst->host, start, len);
    dst->host[len] = '\0';
    dst->ssh->host = xstrdup(dst->host);
    dst->ssh->cmd = NULL;
    dst->ssh->pid = -1;
    start = end + 1;
    end = strpbrk(start, "#/");
    if (end == NULL)
      fatal(1, "Invalid path: %s\n", s);
    len = end - start;
    /* Is port number valid? */
    for(i = 0; i < len; i++) {
      if (!isdigit(start[i]))
        fatal(1, "Invalid port number in path: %s\n", s);
      dst->port = dst->port * 10 + (start[i] - '0');
    };
    start = end + 1;
    if (*end == '#') {
      /* Program name */
      end = strchr(start, '#');
      if (end == NULL || end[1] != '/')
        fatal(1, "Invalid path: %s\n", s);
      len = end - start;
      dst->ssh->cmd = (char *)xmalloc(len + 1);
      strncpy(dst->ssh->cmd, start, len);
      dst->ssh->cmd[len] = '\0';
      start = end + 2;
    }
  } else if (strstr(s, TCP_PREFIX) == s) {
    /* Remote path, process already running */
    start = s + strlen(TCP_PREFIX);
    end = strchr(start, ':');
    if (end == NULL)
      fatal(1, "Invalid path: %s\n", s);
    len = end - start;
    dst->host = (char *)xmalloc(len + 1);
    strncpy(dst->host, start, len);
    dst->host[len] = '\0';
    start = end + 1;
    end = strchr(start, '/');
    if (end == NULL)
      fatal(1, "Invalid path: %s\n", s);
    len = end - start;
    /* Is port number valid? */
    for(i = 0; i < len; i++) {
      if (!isdigit(start[i]))
        fatal(1, "Invalid port number in path: %s\n", s);
      dst->port = dst->port * 10 + (start[i] - '0');
    };
    start = end + 1;
  }
  if (start == s)
    fatal(1, "Invalid path: %s\n", s);
  len = strlen(start);
  if (len == 0 || start[len - 1] == '/')
    fatal(1, "Path must be fully specified (including file name): %s\n", s);
  else {
    dst->path = (char *)xmalloc(len + 1);
    strncpy(dst->path, start, len);
    dst->path[len] = '\0';
  }
}

/* ======================================================================
 * deliver_msg_cb
 */
static int deliver_msg_cb(msg_t *msg, void *arg, int flags)
{
  destination_t *dst = (destination_t *)arg;

  reset_watchdog();

  if (flags == NET_OK) {
    logm(LOG_DEBUG, "%s: Received message from: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);
    if (flux_.protocol->source.deliver(&flux_, msg, dst) < 0) {
      close_destination(dst);
      return -1;
    }
  } else {
    if (flags == NET_DISCONNECT)
      logm(LOG_DEBUG, "%s: Connection closed: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);
    else
      logm(LOG_ERROR, "%s: Error while receiving message from: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);
    flux_.protocol->source.error(&flux_, dst, (flags == NET_DISCONNECT ? ERROR_DISCONNECT : ERROR_RECEIVE));
    close_destination(dst);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * sent_msg_cb
 */
static int sent_msg_cb(msg_t *msg, void *arg, int flags)
{
  destination_t *dst = (destination_t *)arg;

  reset_watchdog();

  if (flags == NET_OK) {
    logm(LOG_DEBUG, "%s: Sent message to: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);
    if (msg->type != MSG_TYPE_INIT) { /* Discard init messages */
      if (flux_.protocol->source.sent(&flux_, msg, dst) < 0) {
        close_destination(dst);
        return -1;
      }
    }
  } else {
    logm(LOG_ERROR, "%s: Error while sending message to: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);
    flux_.protocol->source.error(&flux_, dst, ERROR_SEND);
    close_destination(dst);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * timeout_cb
 */
static int timeout_cb(void *arg)
{
  destination_t *dst = (destination_t *)arg;

  reset_watchdog();

  if (dst == NULL)
    logm(LOG_DEBUG, "%s: Timeout\n", __func__);
  else
    logm(LOG_DEBUG, "%s: Timeout: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);

  if (flux_.protocol->source.timeout(&flux_, dst) < 0 && dst != NULL) {
    close_destination(dst);
    return -1;
  }

  return 0;
}

/* ======================================================================
 * init_source
 */
void init_source(char *src, char **dst, int nb_dst, struct arg_list *ssh_args, config_t *config)
{
  int i;
  char *s;
  char ssh_cmd[512];
  int from_file = 0;

  assert (nb_dst >= 1);

  if (nb_dst == 1 && dst[0][0] == '@') {
    /* Read from file */
    FILE *file;
    char line[512];
    int size = 16;
    char *name = &dst[0][1];

    from_file = 1;
    file = fopen(name, "r");
    if (file == NULL)
      xfatal(1, "%s: Cannot open file: %s\n", __func__, name);

    dst = (char **)xmalloc(size * sizeof(char *));
    nb_dst = 0;

    while (!feof(file)) {
      fgets(line, sizeof(line), file);
      /* Trim spaces */
      s = line + strlen(line) - 1;
      while (s >= line && isspace(*s))
        *(s--) = '\0';
      s = line;
      while (*s != '\0' && isspace(*s))
        *(s++) = '\0';
      if (strlen(s) > 0 && strstr(s, COMMENT_PREFIX) != s) {
        if (nb_dst == size - 1) {
          size *= 2;
          dst = (char **)xrealloc(dst, size * sizeof(char *));
        }
        dst[nb_dst] = xstrdup(s);
        nb_dst++;
      }
    }
    fclose(file);
  }

  parse_source(&flux_, src);
  flux_.file.flux_id = 0;
  s = crypt(flux_.file.file_name, "PF");
  s += 2;
  for (i = 2; i < 13; i++)
    flux_.file.flux_id = (flux_.file.flux_id << 8) | s[i];
  flux_.file.chunk_size = config->chunk_size;
  if (open_file(&flux_.file, OPEN_READ) < 0)
    fatal(1, "%s: Cannot open file: %s\n", __func__, flux_.file.file_name);
  if (flux_.file.file_size == 0)
    fatal(1, "%s: File is empty: %s\n", __func__, flux_.file.file_name);

  flux_.protocol = get_protocol_by_name(config->protocol);
  assert(flux_.protocol != NULL);

  create_stats(&flux_.stats, &flux_.file);
  set_all(&flux_.stats); /* We have all chunks */

  flux_.args = xstrdup(config->args); /* May be modified by source => make a copy */
  flux_.src = NULL;
  flux_.private = NULL;

  flux_.timer.fn = timeout_cb;
  flux_.timer.arg = NULL;
  init_io_timer(&flux_.timer);

  dst_list_ = (destination_t **)xmalloc(nb_dst * sizeof(destination_t *));
  nb_dst_ = 0;
  for (i = 0; i < nb_dst; i++) {
    int fd;
    destination_t *d;

    d = (destination_t *)xmalloc(sizeof(destination_t));

    parse_destination(d, dst[i]);

    if (d->ssh != NULL) {
      snprintf(ssh_cmd, sizeof(ssh_cmd), "%s -d -H %s -P %d",
               d->ssh->cmd != NULL ? d->ssh->cmd : config->program,
               d->host,
               d->port);
      ssh_exec(d->ssh, ssh_args, ssh_cmd);
      if (waitpid(d->ssh->pid, NULL, 0) < 0)
        xfatal(1, "%s: waitpid\n", __func__);
    }

    fd = open_socket(d->host, d->port);
    if (fd < 0) {
      logm(LOG_ERROR, "%s: open_socket\n", __func__);
      free(d);
      continue;
    }

    if (open_io_in(&d->peer.conn.in, fd, deliver_msg_cb, d) < 0) {
      logm(LOG_ERROR, "%s: init_io_in\n", __func__);
      close_socket(fd);
      free(d);
      continue;
    }

    if (open_io_out(&d->peer.conn.out, fd, sent_msg_cb, d) < 0) {
      logm(LOG_ERROR, "%s: init_io_out\n", __func__);
      close_socket(fd);
      close_io_in(&d->peer.conn.in); /* Might try to close socket twice... */
      free(d);
      continue;
    }

    d->peer.private = NULL;

    d->peer.timer.fn = timeout_cb;
    d->peer.timer.arg = d;
    init_io_timer(&d->peer.timer);

    dst_list_[nb_dst_] = d;
    nb_dst_++;
  }
  flux_.nb_peers = nb_dst_;

  if (MUST_LOG(LOG_INFO)) {
    logm(LOG_INFO, "SOURCE\n");
    logm(LOG_INFO, "\tFile ID    : %lu\n", flux_.file.flux_id);
    logm(LOG_INFO, "\tFile name  : %s\n", flux_.file.file_name);
    logm(LOG_INFO, "\tFile size  : %u\n", flux_.file.file_size);
    logm(LOG_INFO, "\tChunk size : %u\n", flux_.file.chunk_size);
    logm(LOG_INFO, "DESTINATION(S)\n");
    for (i = 0; i < nb_dst_; i++) {
      logm(LOG_INFO, "\t%s%s%s:%d%s%s%s/%s\n",
           dst_list_[i]->ssh != NULL && dst_list_[i]->ssh->user != NULL ? dst_list_[i]->ssh->user : "",
           dst_list_[i]->ssh != NULL && dst_list_[i]->ssh->user != NULL ? "@" : "",
           dst_list_[i]->host != NULL ? dst_list_[i]->host : "",
           dst_list_[i]->port,
           dst_list_[i]->ssh != NULL && dst_list_[i]->ssh->cmd != NULL ? "#" : "",
           dst_list_[i]->ssh != NULL && dst_list_[i]->ssh->cmd != NULL ? dst_list_[i]->ssh->cmd : "",
           dst_list_[i]->ssh != NULL && dst_list_[i]->ssh->cmd != NULL ? "#" : "",
           dst_list_[i]->path);
    }
  }

  flux_.protocol->source.init(&flux_, dst_list_, nb_dst_);

  for (i = 0; i < nb_dst_; i++)
    msg_init(&dst_list_[i]->peer, &flux_.file, dst_list_[i]->path, flux_.protocol, flux_.args);

  flux_.protocol->source.start(&flux_);

  if (from_file != 0) {
    for (i = 0; i < nb_dst; i++)
      free(dst[i]);
    free(dst);
  }
}

/* ======================================================================
 * terminate_source
 */
void terminate_source()
{
  int i;

  flux_.protocol->source.terminate(&flux_);

  free(flux_.file.file_name);
  free(flux_.args);
  free_stats(&flux_.stats);

  for (i = 0; i < nb_dst_; i++) {
    close_destination(dst_list_[i]);
    free(dst_list_[i]);
  }
  free(dst_list_);
}

/* ======================================================================
 * close_destination
 */
void close_destination(destination_t *dst)
{
  int i;

  logm(LOG_DEBUG, "%s: Close destination: %s:%d/%s\n", __func__, dst->host, dst->port, dst->path);

  for (i = 0; i < nb_dst_; i++)
    if (dst_list_[i] == dst)
      break;

  assert(i < nb_dst_);

  /* Cleanup */
  close_io_in(&dst->peer.conn.in);
  close_io_out(&dst->peer.conn.out); /* Might try to close socket twice... */
  clear_io_timer(&dst->peer.timer);
  if (dst->ssh != NULL) {
    free(dst->ssh->user);
    free(dst->ssh->host);
    free(dst->ssh->cmd);
    if (dst->ssh->pid > 0)
      kill(dst->ssh->pid, SIGTERM);
    free(dst->ssh);
  }
  free(dst->host);
  free(dst->path);
  free(dst);

  for (; i < nb_dst_ - 1; i++)
    dst_list_[i] = dst_list_[i + 1];
  nb_dst_--;
  flux_.nb_peers--;

  if (flux_.nb_peers == 0) {
    logm(LOG_INFO, "%s: No more active connections\n", __func__);
    logm(LOG_INFO, "%s: Exiting process\n", __func__);
    event_loopexit(NULL);
  }
}
